
void initDb(uint8_t pin, char const *db, char const *cp, uint8_t d) {
  // init DB parameters
  ssP = pin;       // Chip Select pin (D4 is used on my NodeMCU V2 Shield)
  strcpy(dbN, db);  // database file name    //short natation max 8.3 'naaaaame.ext'
  strcpy(cpN, cp);  // copy/backup file name
  dtW = d;          // data width, max of all fields chars and one delimiter/field
  dbW = nxW + dtW;  // record width (nxW is key, 3 is tombstone value + 2 '|')
  dbP = new char[dtW + 1];  // sprintf(), holder fixed data length
  // used in addRec(): create a printf() format string (fixed record size)
  char n[5];
  sprintf(n, "%d", dtW);
  sprintf(dbP, "%%-%ss", n);
  dbOpen();  // SD start, result in dbErr
  if (!dbErr) nrRec();
}

void dbOpen() {
  dbErr = _OK;
  if (dbF) return;
  dbF = SD.open(dbN, FILE_WRITE);
  if (!dbF) dbErr = SD_ERR;
}

void nrRec() {  // set dbR, the nummer of records, tombstone records are excluded
  uint16_t len;
  dbR = 0;
  dbOpen();
  if (dbErr || (dbF.size() < dbW)) return;  // error or no record(s)
  dbF.seek(0);
  while (dbF.available()) {
    len = dbF.readBytes(nxB, nxW);
    nxB[len] = '\0';
    if (len != nxW) {dbErr = CORRUPT; return;}
    if (nxB[tsM] == '0') dbR++; // if it is not a tombstone
    dbF.seek(dbF.position() + dtW); // skip data, set entree to next record
  }
}

boolean findRec(char const *key) {  // slow but simple but fast enough for a small db
  // seek to record, sequentieel search
  // false: not found, true: found
  if (strlen(key) != kyW) {dbErr = MAX_CHARS; return false;}
  dbOpen();
  dbF.seek(0);
  if (dbErr) return false;  // open error
  if (dbF.size() < dbW) {dbErr = DB_EMPTY; return false;}  // db is empty
  uint16_t len;
  while (dbF.available()) {
    len = dbF.readBytes(nxB, nxW);
    nxB[len] = '\0';
    if (len != nxW) {dbErr = CORRUPT; return false;}
    if (nxB[tsM] == '0') {  // not a tombstone?
      boolean f = true;
      for (uint8_t i = 0; i < kyW; i++) {
        if (nxB[i] != key[i]) {f = false; break;}
      }
      if (f) {
        dbF.seek(dbF.position() - nxW);  // seek to the start byte of the record
        return true;  // key is found
      }
    }
    dbF.seek(dbF.position() + dtW);  // skip data, set entree to next record
  }
  dbErr = NOT_FOUND;
  return false;
}

void addRec(char const *key, char const *data) {  // key UNIQUE!!
  if (dbR == mxR) {dbErr = MAX_REC; return;}  // max records in db
  if (strlen(data) > dtW) {dbErr = MAX_CHARS; return;}  // data: too many characters
  boolean f = findRec(key);
  if (f) {dbErr = REC_EXIST; return;}  // record already exists
  if (dbErr != _OK && dbErr != DB_EMPTY && dbErr != NOT_FOUND) return;  // is there a error?
  wrtRec(key, data, true);
}

void writeRec(char const *key, char const *data) {  // key exist (overwrite) or not (add)
  if (strlen(data) > dtW) {dbErr = MAX_CHARS; return;}  // data: too many characters
  boolean f = findRec(key);
  if (dbErr != _OK && dbErr != DB_EMPTY && dbErr != NOT_FOUND) return;  // is there a find error?
  wrtRec(key, data, (f ? false : true));
}

void wrtRec(char const *key, char const *data, boolean eof) {
  dbErr = _OK;  // remove DB_EMPTY of NOT_FOUND
  if (eof) dbF.seek(dbF.size());  // if record not exists then seek to EOF
  // format key and tombstone marker "key|0|" 
  char n[dbW];
  sprintf(n, "%3s|0|", key);
  dbF << n;  // write to db
  // format data with trailing white spaces
  sprintf(n, dbP, data);
  dbF << n;  // write to db
  dbF.flush();  // save record physically to the SD
  if (eof) dbR++;  // update record nr
}

void readRec(char const *key) {  // read a record, it is in record buffer dbB
  if (!findRec(key)) return;  // find position in file
  delete dbB;
  dbB = new char[dbW + 1];  // record read buffer)
  uint8_t len = dbF.readBytes(dbB, dbW);  // record in buffer dbB
  dbB[len] = '\0';
  if (len != dbW) {dbErr = CORRUPT; return;}
}

void delRec(char const *key) {  // tombstone a record, remove tombstones with comprDb
  if (!findRec(key)) return;  // find position in file
  dbF.seek(dbF.position() + 4); 
  dbF << '1';
  dbF.flush();
  dbR--;  // update record nr
}

void listRec(boolean data) {
  uint16_t len;
  uint16_t n = 0;
  dbOpen();
  if (dbErr) return;
  if (dbF.size() < dbW) {dbErr = DB_EMPTY; return;}  // no record(s)
  dbF.seek(0);
  while (dbF.available()) {
    len = dbF.readBytes(nxB, nxW);
    nxB[len] = '\0';
    if (len != nxW) {dbErr = CORRUPT; return;}
    if (nxB[tsM] == '0') { // if it is not a tombstone
      nxB[kyW] = '\0';
      Serial << n++ << ": " << nxB;
      if (data) {
        delete dbB;
        dbB = new char[dbW + 1];  // record read buffer)
        len = dbF.readBytes(dbB, dtW);
        dbB[len] = '\0';
        Serial << " - " << dbB;
     }
     else dbF.seek(dbF.position() + dtW); // skip data, set entree to next record
     Serial << endl;
    }
  }
}

void cmprDb() {  // copy db, then remove tombstone records
  // there in no rename() function in Arduino SD library :( 
  if (SD.exists(cpN)) if (!SD.remove(cpN)) {dbErr = REMOVE_ERR; return;}  // delete file error
  cpF = SD.open(cpN, FILE_WRITE);
  if (!cpF) {dbErr = SD_ERR; return;}
  dbOpen();
  if (dbErr) return;

  // make a copy
  dbF.seek(0);
  while (dbF.available()) {
    cpF << (char)dbF.read();
  }
  cpF.flush();
  dbF.close();
  // delete db
  if (!SD.remove(dbN)) {dbErr = REMOVE_ERR; return;}  // delete file error

  // make a new db file
  dbOpen();
  if (dbErr) return;
  if (dbF.size() != 0) {dbErr = NEW_ERR; return;}  // file not earased - is not possible?

  // copy the 'not as tombstone marked' records from copy to database
  uint16_t len;
  int16_t nr = 0;
  cpF.seek(0);
  while (cpF.available()) {
   delete dbB;
   dbB = new char[dbW + 1];  // record read buffer)
   len = cpF.readBytes(dbB, dbW);
    dbB[len] = '\0';
    if (strlen(dbB) != len) {
      cpF.close();
      dbErr = CORRUPT;
      return;  // db corrupt
    }
    if (dbB[tsM] == '0') {  // if it is not a tombstone ...
      dbF << dbB;  // ... then copy it to the tmp file
      dbF.flush();
      nr++;
    }
  }
  cpF.close();  // do not remove, it is your backup
  dbR = nr;
  if (nr != dbR) {dbErr = DIF_NR; return;}  // difference in number of records
  dbErr = _OK;
}

void showErr() {
  char s[32];
  switch (dbErr) {
    case _OK: return; break;
    case SD_ERR:     strcpy(s, "sd error"); break;
    case DB_EMPTY:   strcpy(s, "no record(s) in db"); break;
    case MAX_REC:    strcpy(s, "max records in db"); break;
    case MAX_CHARS:  strcpy(s, "key or data: too many characters"); break;
    case REC_EXIST:  strcpy(s, "record already exists"); break;
    case NOT_FOUND:  strcpy(s, "record not found"); break;
    case REMOVE_ERR: strcpy(s, "delete file error"); break;
    case CORRUPT:    strcpy(s, "db corrupt?"); break;
    case DIF_NR:     strcpy(s, "difference in number of records"); break;
    case NEW_ERR:    strcpy(s, "file not earased"); break;
  }
  Serial << "\t### " << s << endl;
}

